@@ -46,10 +46,6 @@ extern char _STR[150];
 extern short fontcount;
 extern fontinfo *fonts;
 
-/*------------------------------------------------------*/
-/* Check if a library exists				*/
-/*-------------------------------------------------------*/
-
 /*---------------------------------------------------------*/
 /* Find the Helvetica font for use in labeling the objects */
 /*---------------------------------------------------------*/
@@ -249,6 +245,23 @@ short *pagetotals(int page, short mode)
    }
 }
 
+/*---------------------------------------------------------*/
+/* Test whether a library instance is a "virtual" instance */
+/*---------------------------------------------------------*/
+
+Boolean is_virtual(objinstptr thisinst) {
+   int libno;
+   liblistptr ilist;
+
+   libno = libfindobject(thisinst->thisobject, NULL);
+
+   for (ilist = xobjs.userlibs[libno].instlist; ilist != NULL; ilist = ilist->next)
+      if ((ilist->thisinst == thisinst) && (ilist->virtual == TRUE))
+	 return TRUE;
+
+   return FALSE;
+}
+
 /*------------------------------------------------------*/
 /* Test whether an object is a page, and return the	*/
 /* page number if it is.  Otherwise, return -1.		*/
@@ -296,6 +309,95 @@ int check_library(char *libname)
    return -1;
 }
 
+/*------------------------------------------------------*/
+/* Move an object and all of its virtual instances from	*/
+/* one library to another.				*/
+/*------------------------------------------------------*/
+
+int libmoveobject(objectptr thisobject, int libtarget)
+{
+   int j, libsource;
+   liblistptr spec, slast, srch;
+
+   libsource = libfindobject(thisobject, &j);
+
+   if (libsource == libtarget) return libsource; /* nothing to do */
+   if (libsource < 0) return libsource;		 /* object not in the library */
+
+   /* Move the object from library "libsource" to library "libtarget" */
+
+   xobjs.userlibs[libtarget].library = (objectptr *)
+		realloc(xobjs.userlibs[libtarget].library,
+		(xobjs.userlibs[libtarget].number + 1) * sizeof(objectptr));
+
+   *(xobjs.userlibs[libtarget].library + xobjs.userlibs[libtarget].number) = thisobject;
+   xobjs.userlibs[libtarget].number++;
+
+   for (; j < xobjs.userlibs[libsource].number; j++)
+      *(xobjs.userlibs[libsource].library + j) =
+		*(xobjs.userlibs[libsource].library + j + 1);
+      xobjs.userlibs[libsource].number--;
+
+   /* Move all instances from library "libsource" to library "libtarget" */
+
+   slast = NULL;
+   for (spec = xobjs.userlibs[libsource].instlist; spec != NULL;) {
+      if (spec->thisinst->thisobject == thisobject) {
+
+	 /* Add to end of spec list in target */
+	 srch = xobjs.userlibs[libtarget].instlist;
+	 if (srch == NULL)
+	    xobjs.userlibs[libtarget].instlist = spec;
+	 else {
+	    for (; srch->next != NULL; srch = srch->next);
+	    spec->next = srch->next;
+	    srch->next = spec;
+	 }
+	 
+	 if (slast != NULL) {
+	    slast->next = spec->next;
+	    spec = slast->next;
+	 }
+	 else {
+	    xobjs.userlibs[libsource].instlist = spec->next;
+	    spec = xobjs.userlibs[libsource].instlist;
+	 }
+      }
+      else {
+	 slast = spec;
+	 spec = spec->next;
+      }
+   }
+
+   return libsource;
+}
+
+/*------------------------------------------------------*/
+/* Determine which library contains the specified	*/
+/* object.  If found, return the library number, or	*/
+/* else return -1 if the object was not found in any	*/
+/* library.  If "partidx" is non-null, fill with the	*/
+/* integer offset of the object from the beginning of	*/
+/* the library.						*/
+/*------------------------------------------------------*/
+
+int libfindobject(objectptr thisobject, int *partidx)
+{
+   int i, j;
+   objectptr libobj;
+
+   for (i = 0; i < xobjs.numlibs; i++) {
+      for (j = 0; j < xobjs.userlibs[i].number; j++) {
+         libobj = *(xobjs.userlibs[i].library + j);
+	 if (libobj == thisobject) {
+	    if (partidx != NULL) *partidx = j;
+	    return i;
+	 }
+      }
+   }
+   return -1;
+}
+
 /*------------------------------------------------------*/
 /* ButtonPress handler during page catalog viewing mode */
 /*------------------------------------------------------*/
@@ -781,7 +883,6 @@ void composelib(short mode)
 
    /* Compute the bounding box of the library page */
    calcbbox(xobjs.libtop[mode]);
-   centerview(xobjs.libtop[mode]);
 
    /* Update the library directory */
    updatepagelib(LIBLIB, mode);
@@ -852,17 +953,8 @@ void catvirtualcopy()
    for (newselect = areawin->selectlist; newselect < areawin->selectlist
 	  + areawin->selects; newselect++) {
       libobj = SELTOOBJINST(newselect);
-
-      if (libobj->thisobject->params == NULL) {
-	 Wprintf("Virtual copy allowed only on objects with paramters.");
-	 /* (because currently there's no way to adjust rotation/scale	*/
-	 /* of a non-parameterized instance from inside the library,	*/
-	 /* even though it can be specified in the library itself)	*/
-      }
-      else { 		/* All's clear to make a virtual copy */
-	 libinst = addtoinstlist(i, libobj->thisobject, TRUE);
-	 instcopy(libinst, libobj);
-      }
+      libinst = addtoinstlist(i, libobj->thisobject, TRUE);
+      instcopy(libinst, libobj);
    }
 
    clearselects();
@@ -1084,7 +1176,7 @@ void linkedlistinsertafter(liblistptr *spec, int o1, int o2)
 
 void catmove(int x, int y)
 {
-   int i, j, s1, s2, ocentx, ocenty, rangex, rangey;
+   int i, j, k, s1, s2, ocentx, ocenty, rangex, rangey, l;
    liblistptr spec;
    objinstptr exchobj, lobj;
 
@@ -1096,79 +1188,89 @@ void catmove(int x, int y)
    }
 
    if (areawin->selects == 0) return;
-   else if (areawin->selects > 2) {
-      Wprintf("Select maximum of two objects.");
-      return;
-   }
 
-   exchobj = SELTOOBJINST(areawin->selectlist);
-   s1 = -1;
+   /* Add selected object or objects at the cursor position */
+
+   window_to_user(x, y, &areawin->save);
+
+   s2 = -1;
    for (j = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
 		spec = spec->next, j++) {
-      if (spec->thisinst == exchobj) {
-	 s1 = j;
-	 break;
+      lobj = spec->thisinst;
+      for (k = 0; k < areawin->selects; k++) {
+	 exchobj = SELTOOBJINST(areawin->selectlist + k);
+	 if (lobj == exchobj) break;
       }
-   }
+      if (k < areawin->selects) continue;	/* ignore items being moved */
 
-   /* If two objects are selected, then exchange their order */
-
-   if (areawin->selects == 2) {
-      lobj = SELTOOBJINST(areawin->selectlist + 1);
-
-      s2 = -1;
-      for (j = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
-		spec = spec->next, j++) {
-	 if (spec->thisinst == lobj) {
-	    s2 = j;
-	    break;
-	 }
-      }
-      /* Exchange numbers s1 and s2 */
-      linkedlistswap(&(xobjs.userlibs[i].instlist), s1, s2);
-   }
-   else {	/* one object selected; find place to put from cursor position */
-
-      window_to_user(x, y, &areawin->save);
-
-      s2 = -1;
-      for (j = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
-		spec = spec->next, j++) {
-	 lobj = spec->thisinst;
-
- 	 ocentx = lobj->position.x + lobj->bbox.lowerleft.x
+      ocentx = lobj->position.x + lobj->bbox.lowerleft.x
 	        + (lobj->bbox.width >> 1);
- 	 ocenty = lobj->position.y + lobj->bbox.lowerleft.y
+      ocenty = lobj->position.y + lobj->bbox.lowerleft.y
 	        + (lobj->bbox.height >> 1);
-	 rangey = (lobj->bbox.height > 200) ? 
+      rangey = (lobj->bbox.height > 200) ? 
 	        (lobj->bbox.height >> 1) : 100;
 
-	 if ((areawin->save.y < ocenty + rangey) && (areawin->save.y 
+      if ((areawin->save.y < ocenty + rangey) && (areawin->save.y 
 	         > ocenty - rangey)) {
-	    s2 = j - 1;
-	    if (areawin->save.x < ocentx) break;
-	    else s2 = j;
-	 }
+	 s2 = j - 1;
+	 if (areawin->save.x < ocentx) break;
+	 else s2 = j;
       }
-      if ((s2 == -1) && (spec == NULL)) {
-	 if (areawin->save.y <
+   }
+   if ((s2 == -1) && (spec == NULL)) {
+      if (areawin->save.y <
 		xobjs.libtop[i + LIBRARY]->thisobject->bbox.lowerleft.y)
-	    s2 = j - 1;
-	 else if (areawin->save.y <=
+	 s2 = j - 1;
+      else if (areawin->save.y <=
 		xobjs.libtop[i + LIBRARY]->thisobject->bbox.lowerleft.y +
 		xobjs.libtop[i + LIBRARY]->thisobject->bbox.height) {
-	    unselect_all();	
-	    Wprintf("Could not find appropriate place to insert object");
-	    return;
-	 }
+	 unselect_all();	
+	 Wprintf("Could not find appropriate place to insert object");
+	 return;
       }
+   }
 
-      /* Move number s1, insert it after s2 */
-      linkedlistinsertafter(&(xobjs.userlibs[i].instlist), s1, s2);
+   /* Find object number s2 (because s2 value may change during insertion) */
+   if (s2 > -1) {
+      for (k = 0, spec = xobjs.userlibs[i].instlist; k < s2; spec = spec->next, k++);
+      lobj = spec->thisinst;
+   }
+   else lobj = NULL;
+
+   /* Move items; insert them after item s2 in order selected */
+
+   j = i;
+   for (k = 0; k < areawin->selects; k++) {
+
+      /* Find number of lobj (may have changed) */
+      for (s2 = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
+		spec = spec->next, s2++)
+	 if (spec->thisinst == lobj)
+	    break;
+
+      exchobj = SELTOOBJINST(areawin->selectlist + k);
+      for (s1 = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
+		spec = spec->next, s1++)
+         if (spec->thisinst == exchobj)
+	    break;
+
+      if (spec == NULL) {
+	 /* Object came from another library */
+         if ((l = libmoveobject(exchobj->thisobject, i)) >= 0) j = l;
+      }
+      else {
+         linkedlistinsertafter(&(xobjs.userlibs[i].instlist), s1, s2);
+      }
    }
 
    unselect_all();
-   if ((i = is_library(topobject)) >= 0) composelib(LIBRARY + i);
+   composelib(LIBRARY + i);
+   incr_changes(xobjs.libtop[LIBRARY + i]->thisobject);
+   if (j != i) {
+      composelib(LIBRARY + j);
+      centerview(xobjs.libtop[LIBRARY + j]);
+      incr_changes(xobjs.libtop[LIBRARY + j]->thisobject);
+   }
 
    drawarea(NULL, NULL, NULL);
 }
@@ -1405,8 +1507,8 @@ void catalog_op(int op, int x, int y)
 	       if (eventmode == ASSOC_MODE) {
 
 	          /* revert to old page */
-		  topobject->viewscale =areawin->vscale;
-		  topobject->pcorner =areawin->pcorner;
+		  topobject->viewscale = areawin->vscale;
+		  topobject->pcorner = areawin->pcorner;
 	          areawin->topinstance = (areawin->stack == NULL) ?
 			   xobjs.pagelist[areawin->page]->pageinst
 			   : areawin->stack->thisinst;
@@ -1420,8 +1522,8 @@ void catalog_op(int op, int x, int y)
 
 	          /* revert to old page */
 
-		  topobject->viewscale =areawin->vscale;
-		  topobject->pcorner =areawin->pcorner;
+		  topobject->viewscale = areawin->vscale;
+		  topobject->pcorner = areawin->pcorner;
 	          areawin->topinstance = (areawin->stack == NULL) ?
 			   xobjs.pagelist[areawin->page]->pageinst
 			   : areawin->stack->thisinst;
@@ -1544,6 +1646,10 @@ void changecat()
       }
       areawin->lastlibrary = j;
    }
+
+   if (eventmode == CATMOVE_MODE)
+      delete_for_xfer(NORMAL, areawin->selectlist, areawin->selects);
+
    startcatalog(NULL, (pointertype)(j + LIBRARY), NULL);
 }
 
@@ -1572,7 +1678,7 @@ void startcatalog(xcWidget w, pointertype libmod, caddr_t nulldata)
       XDefineCursor (dpy, areawin->window, DEFAULTCURSOR);
       eventmode = CATALOG_MODE;
    }
-   else
+   else if (eventmode != CATMOVE_MODE)
       eventmode = CATALOG_MODE;
 
    /* Push the current page onto the push stack, unless	we're going */
@@ -1589,12 +1695,14 @@ void startcatalog(xcWidget w, pointertype libmod, caddr_t nulldata)
 
    /* set library as new object */
 
-   topobject->viewscale =areawin->vscale;
-   topobject->pcorner =areawin->pcorner;
+   topobject->viewscale = areawin->vscale;
+   topobject->pcorner = areawin->pcorner;
    areawin->topinstance = xobjs.libtop[libmod];
    setpage(TRUE);
    invalidate_graphics(topobject);
 
+   transferselects();
+
    /* draw the new screen */
 
    refresh(NULL, NULL, NULL);
