@@ -159,17 +159,17 @@ extern XPoint *pathclosepoint(pathptr, XPoint *);
 extern void placeselects(short, short, XPoint *);
 extern void drag(int, int);
 extern void xlib_drag(xcWidget, caddr_t, XEvent *);
-extern void elemrotate(genericptr *, short);
-extern void elementrotate(short);
+extern void elemrotate(genericptr *, short, int, int);
+extern void elementrotate(short, int, int);
 extern void edit(int, int);
 extern void pathedit(genericptr *, short);
 extern void xc_lower(short *);
 extern void xc_raise(short *);
 extern void exchange();
-extern void elhflip(genericptr *);
-extern void elvflip(genericptr *);
-extern void elementflip();
-extern void elementvflip();
+extern void elhflip(genericptr *, int, int);
+extern void elvflip(genericptr *, int, int);
+extern void elementflip(int, int);
+extern void elementvflip(int, int);
 #ifdef TCL_WRAPPER
 extern xcTimeOutProc makepress(caddr_t);
 #else
@@ -412,6 +412,7 @@ extern int string_to_func(const char *, short *);
 extern int string_to_key(const char *);
 extern char *function_binding_to_string(xcWidget, int);
 extern char *key_binding_to_string(xcWidget, int);
+extern char *compat_key_to_string(xcWidget, int);
 extern char *func_to_string(int);
 extern char *key_to_string(int);
 extern void printeditbindings();
@@ -423,8 +424,8 @@ extern int remove_binding(xcWidget, int, int);
 extern void remove_keybinding(xcWidget, const char *, const char *);
 
 #ifndef TCL_WRAPPER
-extern void mode_rebinding(xcWidget, int, int);
-extern void mode_tempbinding(xcWidget, int, int);
+extern void mode_rebinding(int, int);
+extern void mode_tempbinding(int, int);
 #endif
 
 /* from libraries.c: */
@@ -434,8 +435,11 @@ extern void catreturn();
 extern int pageposition(short, int, int, int);
 extern short pagelinks();
 extern short *pagetotals(int, short);
+extern Boolean is_virtual(objinstptr);
 extern int is_page(objectptr);
 extern int is_library(objectptr);
+extern int libfindobject(objectptr, int *);
+extern int libmoveobject(objectptr, int);
 extern void pagecat_op(int, int, int);
 extern void pageinstpos(short, short, objinstptr, int, int, int, int);
 extern objinstptr newpageinst(objectptr);
@@ -637,6 +641,9 @@ extern void destroynets(objectptr);
 extern int  cleartraversed(objectptr);
 extern int  checkvalid(objectptr);
 extern void clearlocalpins(objectptr);
+extern void append_included(char *);
+extern Boolean check_included(char *);
+extern void free_included();
 
 /* from parameter.c: */ 
 
@@ -749,7 +756,6 @@ extern Boolean render_client();
 extern objectptr NameToPageObject(char *, objinstptr *, int *);
 extern objectptr NameToObject(char *, objinstptr *, Boolean);
 extern void callwritenet(xcWidget, pointertype, caddr_t);
-extern void doxschema(xcWidget, pointertype, caddr_t);
 extern void startconnect(xcWidget, caddr_t, caddr_t);
 extern void connectivity(xcWidget, caddr_t, caddr_t);
 extern Boolean setobjecttype(objectptr);
